import os
import itertools
import numpy as np
import matplotlib.pyplot as plt

from sklearn.preprocessing import MinMaxScaler
from sklearn.datasets import make_regression

# Trees and forests
from sklearn.tree import DecisionTreeRegressor
from sklearn.ensemble import RandomForestRegressor, AdaBoostRegressor

# KNN
from sklearn.neighbors import KNeighborsRegressor

# Linear regression
from sklearn.linear_model import LinearRegression

# MLPs
from sklearn.neural_network import MLPRegressor

# Polynomial and kernel
from sklearn.kernel_ridge import KernelRidge

from sklearn.inspection import DecisionBoundaryDisplay
import matplotlib.font_manager as fm

myfont = fm.FontProperties(fname="./FiraSans-Regular.ttf")


def sqdists(x, y):
    return np.sum((x[:, None, :] - y[None, :, :]) ** 2, axis=-1)


def kernel_matrix(x, y, *, kernel, gamma):
    if kernel == "cauchy":
        return 1 / (1 + gamma * sqdists(x, y))
    elif kernel == "rbf":
        return np.exp(-gamma * sqdists(x, y))
    elif kernel == "laplace":
        return np.exp(-gamma * np.sqrt(sqdists(x, y)))
    elif kernel == "distance" and gamma == 1:
        return -np.sqrt(sqdists(x, y))
    elif kernel == "distance" and gamma == -1 / 2:
        return 1 / (np.sqrt(np.sqrt(sqdists(x, y))) + 1e-8)
    elif kernel == "distance" and gamma == -1:
        return 1 / (np.sqrt(sqdists(x, y)) + 1e-8)
    elif kernel == "distance" and gamma == -2:
        return 1 / (sqdists(x, y) + 1e-8)
    elif kernel == "distance" and gamma == -4:
        return 1 / (sqdists(x, y) ** 2 + 1e-8)


class MyKernelRidge:
    def __init__(self, *, alpha, kernel, gamma=None):
        self.alpha = alpha
        self.kernel = kernel
        self.gamma = gamma

    def fit(self, X, y):
        inner = KernelRidge(alpha=self.alpha, kernel="precomputed")
        K = kernel_matrix(X, X, kernel=self.kernel, gamma=self.gamma)
        inner.fit(K, y)

        self.X = X
        self.dual_coef_ = inner.dual_coef_

    def predict(self, X):
        K = kernel_matrix(X, self.X, kernel=self.kernel, gamma=self.gamma)
        return K @ self.dual_coef_


class NWKernel:
    def __init__(self, *, kernel, gamma=None):
        self.kernel = kernel
        self.gamma = gamma

    def fit(self, X, y):
        self.X_ = X
        self.y_ = y

    def predict(self, X):
        K = kernel_matrix(X, self.X_, kernel=self.kernel, gamma=self.gamma)
        return (K @ self.y_) / np.sum(K, axis=1)


mlp_opt = {
    "solver": "lbfgs",
    "learning_rate_init": 0.1,
    "max_iter": 1000,
    "alpha": 0,
    "random_state": 421,
}

models_tree_knn = [
    # Raw dataset:
    ("dataset", None),
    # Simple trees:
    ("decision_tree_1", DecisionTreeRegressor(max_depth=1)),
    ("decision_tree_2", DecisionTreeRegressor(max_depth=2)),
    ("decision_tree_3", DecisionTreeRegressor(max_depth=3)),
    ("decision_tree_4", DecisionTreeRegressor(max_depth=4)),
    ("decision_tree_5", DecisionTreeRegressor(max_depth=5)),
    ("decision_tree_10", DecisionTreeRegressor(max_depth=10)),
    # Random forests:
    ("forest_3_1", RandomForestRegressor(max_depth=3, n_estimators=1)),
    ("forest_3_2", RandomForestRegressor(max_depth=3, n_estimators=2)),
    ("forest_3_5", RandomForestRegressor(max_depth=3, n_estimators=5)),
    ("forest_3_100", RandomForestRegressor(max_depth=3, n_estimators=100)),
    # Tree bosting:
    (
        "adaboost_3_1",
        AdaBoostRegressor(DecisionTreeRegressor(max_depth=3), n_estimators=1),
    ),
    (
        "adaboost_3_3",
        AdaBoostRegressor(DecisionTreeRegressor(max_depth=3), n_estimators=3),
    ),
    (
        "adaboost_3_5",
        AdaBoostRegressor(DecisionTreeRegressor(max_depth=3), n_estimators=5),
    ),
    (
        "adaboost_3_100",
        AdaBoostRegressor(DecisionTreeRegressor(max_depth=3), n_estimators=100),
    ),
    # K nearest neighbors:
    ("knn_1", KNeighborsRegressor(n_neighbors=1)),
    ("knn_2", KNeighborsRegressor(n_neighbors=2)),
    ("knn_3", KNeighborsRegressor(n_neighbors=3)),
    ("knn_4", KNeighborsRegressor(n_neighbors=4)),
    ("knn_5", KNeighborsRegressor(n_neighbors=5)),
    # Linear regression:
    ("linear", LinearRegression()),
]

models_mlp = [
    # Neural networks:
    ("mlp_1", MLPRegressor((1,), **mlp_opt)),
    ("mlp_10", MLPRegressor((10,), **mlp_opt)),
    ("mlp_20", MLPRegressor((20,), **mlp_opt)),
    ("mlp_50", MLPRegressor((50,), **mlp_opt)),
    ("mlp_100", MLPRegressor((100,), **mlp_opt)),
    ("mlp_100_100", MLPRegressor((100, 100), **mlp_opt)),
    ("mlp_100_100_100", MLPRegressor((100, 100, 100), **mlp_opt)),
    ("mlp_100_100_100_100", MLPRegressor((100, 100, 100, 100), **mlp_opt)),
]

models_kernel = [
    # Polynomial interpolation:
    ("kernel_polynomial_d0", KernelRidge(alpha=0, kernel="polynomial", degree=0)),
    ("kernel_polynomial_d1", KernelRidge(alpha=0, kernel="polynomial", degree=1)),
    ("kernel_polynomial_d2", KernelRidge(alpha=0, kernel="polynomial", degree=2)),
    ("kernel_polynomial_d3", KernelRidge(alpha=0, kernel="polynomial", degree=3)),
    ("kernel_polynomial_d4", KernelRidge(alpha=0, kernel="polynomial", degree=4)),
    ("kernel_polynomial_d5", KernelRidge(alpha=0, kernel="polynomial", degree=5)),
    ("kernel_polynomial_d10", KernelRidge(alpha=0, kernel="polynomial", degree=10)),
    ("kernel_polynomial_d20", KernelRidge(alpha=0, kernel="polynomial", degree=20)),
    # Gaussian:
    ("kernel_gaussian_s1", KernelRidge(alpha=0, kernel="rbf", gamma=1 / (2 * 1**2))),
    # With ridge regularization:
    (
        "kernel_gaussian_a01_s1",
        KernelRidge(alpha=0.1, kernel="rbf", gamma=1 / (2 * 1**2)),
    ),
    (
        "kernel_gaussian_a01_s05",
        KernelRidge(alpha=0.1, kernel="rbf", gamma=1 / (2 * 0.5**2)),
    ),
    (
        "kernel_gaussian_a01_s02",
        KernelRidge(alpha=0.1, kernel="rbf", gamma=1 / (2 * 0.2**2)),
    ),
    (
        "kernel_gaussian_a01_s01",
        KernelRidge(alpha=0.1, kernel="rbf", gamma=1 / (2 * 0.1**2)),
    ),
    # Cauchy:
    (
        "kernel_cauchy_a01_s02",
        MyKernelRidge(alpha=0.1, kernel="cauchy", gamma=1 / 0.2**2),
    ),
    # Laplace:
    (
        "kernel_laplace_a01_s02",
        MyKernelRidge(alpha=0.1, kernel="laplace", gamma=1 / 0.2),
    ),
    (
        "kernel_laplace_a01_s05",
        MyKernelRidge(alpha=0.1, kernel="laplace", gamma=1 / 0.5),
    ),
    # ED:
    (
        "kernel_distance_a01_g1",
        MyKernelRidge(alpha=0.1, kernel="distance", gamma=1),
    ),
]


models_nadaraya = [
    ("nw_distance_g1", NWKernel(kernel="distance", gamma=1)),
    ("nw_distance_g-05", NWKernel(kernel="distance", gamma=-1 / 2)),
    ("nw_distance_g-1", NWKernel(kernel="distance", gamma=-1)),
    ("nw_distance_g-2", NWKernel(kernel="distance", gamma=-2)),
    ("nw_distance_g-4", NWKernel(kernel="distance", gamma=-4)),
    ("nw_gaussian_g01", NWKernel(kernel="rbf", gamma=1 / 0.1**2)),
    ("nw_gaussian_g02", NWKernel(kernel="rbf", gamma=1 / 0.2**2)),
    ("nw_laplace_g01", NWKernel(kernel="laplace", gamma=1 / 0.1)),
    ("nw_laplace_g02", NWKernel(kernel="laplace", gamma=1 / 0.2)),
    ("nw_cauchy_g01", NWKernel(kernel="cauchy", gamma=1 / 0.1**2)),
    ("nw_cauchy_g02", NWKernel(kernel="cauchy", gamma=1 / 0.2**2)),
]


models = models_tree_knn + models_mlp + models_kernel + models_nadaraya
# models = models_nadaraya


datasets = [
    ("2D_n100", make_regression(n_samples=20, n_features=2, noise=100, random_state=0)),
]

datasets = [
    (
        ds_name,
        (
            MinMaxScaler((-0.4, 0.4)).fit_transform(X),
            MinMaxScaler((-1, 1)).fit_transform(y.reshape(-1, 1)).reshape(-1),
        ),
    )
    for ds_name, (X, y) in datasets
]


X_1D = np.array([1.0, 1.4, 2.0, 2.3, 2.9, 4.0, 4.5, 5.0, 5.4]).reshape(-1, 1) / 6 - 0.5
y_1D = np.array([1.5, 0.8, 0.7, 0.2, 0.42, 1.0, 1.4, 1.1, 1.9]) / 6 - 0.25

datasets = [
    ("1D_bowl", (X_1D, y_1D)),
] + datasets

print(datasets)


# iterate over datasets

for ds_name, (X, y) in datasets:
    print(ds_name, "=====================================")
    D = X.shape[1]

    if D == 1:
        figure = plt.figure(figsize=(8, 5), dpi=300, facecolor="#fafafa")

    else:
        figure = plt.figure(figsize=(12, 12), facecolor="#fafafa")
        cm = plt.cm.RdBu
        vmin = -1.2
        vmax = 1.2

    for name, model in models:
        print(name, end=": ", flush=True)
        if model is not None:
            model.fit(X, y)
            pred = model.predict(X)
            print(np.mean(np.abs(pred - y)))

        plt.clf()
        ax = plt.gca()
        ax.set_facecolor("#fafafa")

        if D == 1:
            T = np.linspace(-0.5, 0.5, 10001).reshape(-1, 1)
            if model is not None:
                u = model.predict(T)

                segs_x = [[a, a, None] for a in X.reshape(-1)]
                segs_x = np.array(list(itertools.chain.from_iterable(segs_x)))
                segs_y = [[b, c, None] for b, c in zip(y, pred)]
                segs_y = np.array(list(itertools.chain.from_iterable(segs_y)))

            # get width and height of axes object to compute
            # matching arrowhead length and width
            dps = plt.gcf().dpi_scale_trans.inverted()
            bbox = ax.get_window_extent().transformed(dps)
            width, height = bbox.width, bbox.height

            xmin, xmax = -0.5, 0.5
            ymin, ymax = -0.3, 0.2
            # manual arrowhead width and length
            hw = 1.0 / 20.0 * (ymax - ymin)
            hl = 1.0 / 20.0 * (xmax - xmin)
            lw = 1.0  # axis line width
            ohg = 0.3  # arrow overhang

            # compute matching arrowhead length and width
            yhw = hw / (ymax - ymin) * (xmax - xmin) * height / width
            yhl = hl / (xmax - xmin) * (ymax - ymin) * width / height

            # draw x and y axis
            ax.arrow(
                xmin,
                -0.25,
                xmax - xmin,
                0.0,
                fc="k",
                ec="k",
                lw=lw,
                head_width=hw,
                head_length=hl,
                overhang=ohg,
                length_includes_head=True,
                clip_on=False,
            )

            ax.arrow(
                -0.45,
                ymin,
                0.0,
                ymax - ymin,
                fc="k",
                ec="k",
                lw=lw,
                head_width=.8*yhw,
                head_length=yhl,
                overhang=ohg,
                length_includes_head=True,
                clip_on=False,
            )

            ax.text(
                0.45,
                -0.3,
                r"x",
                fontproperties=myfont,
                fontsize=35,
                color="#080899",
            )
            ax.text(
                -0.495,
                0.15,
                r"y",
                fontproperties=myfont,
                fontsize=35,
                color="#005F5F", #"#990808",
            )

            if model is not None:
                ax.scatter(
                    X.reshape(-1),
                    pred,
                    40, #160,
                    color="b",
                    linewidths=1,
                    edgecolors="#000080",
                )
                ax.plot(T.reshape(-1), u, color="b", linewidth=2)
                ax.plot(segs_x, segs_y, color="#95DFEB", linewidth=4, zorder=0)

            ax.scatter(
                X.reshape(-1), y, 160, color="c", linewidths=2, edgecolors="#005F5F", zorder=4
            )

            ax.axis([-0.5, 0.5, -0.3, 0.2])
            ax.set_aspect("equal")
            ax.axis("off")

        elif D == 2:
            if model is not None:
                DecisionBoundaryDisplay.from_estimator(
                    model,
                    np.array([[-0.5, -0.5], [0.5, 0.5]]),
                    cmap=cm,
                    alpha=0.8,
                    ax=ax,
                    eps=0,
                    plot_method="pcolormesh",
                    grid_resolution=1000,
                    vmin=vmin,
                    vmax=vmax,
                )

            # Plot the training points
            ax.scatter(
                X[:, 0],
                X[:, 1],
                s=25**2,
                c=y,
                cmap=cm,
                linewidths=5,
                edgecolors="k",
                vmin=vmin,
                vmax=vmax,
            )

            ax.set_xlim(-0.5, 0.5)
            ax.set_ylim(-0.5, 0.5)
            ax.set_xticks(())
            ax.set_yticks(())

        plt.tight_layout()
        plt.show()

        filename = f"output_regression/{ds_name}/{name}.png"
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        plt.savefig(filename, bbox_inches="tight", transparent=False)
